# -*- tcl -*-
# $Id: sandlib.tcl,v 1.141 2003/05/22 15:45:47 lioka Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package require Trf
package require Memchan
package require cmdline
package provide sandlib 0.1

namespace eval sandlib {
    # config dir
    array set layout \
	[list \
	     etcdir /etc/sandman \
	     libdir [file dirname [info script]] \
	     prefix /var/lib/sandman \
	     logdir /var/log/sandman \
	     cachedir /var/cache/sandman \
	     ccachedir /var/cache/sandman/ccache \
	     tmpdir /tmp]

    variable prog \
	[list \
	     [list sandd \
		  [list \
		       [list modules {pm box cvs rep task serv}] \
		       [list init [list rep server]]]] \
	     [list sandctl \
		  [list \
		       [list modules tool] \
		       [list init tool]]] \
	     [list sandcl \
		  [list \
		       [list modules cl] \
		       [list init client]]]]

    array set log [list header {} msg stdout dbg stderr err stderr]
    variable onerror errorResult

    namespace export log parse perror popen chroot \
	safejoin readlink symlink fileq
}

###############################################################################
# init everything
#
proc sandlib::init {} {
    variable layout
    variable prog

    set progname [file tail $::argv0]

    if {![keylget prog $progname proglist]} {
	puts stderr "$::argv0: i'm little green man"
	exit 1
    }

    if {![id userid]} {
	puts stderr "do not run as root please"
	exit 1
    }

    #umask 7

    try_eval {
	foreach f [keylget proglist modules] {
	    source [file join $layout(libdir) sand$f.tcl]
	}
	foreach p [keylget proglist init] {
	    ${p}::init
	}
    } {
	log err [perror]
	exit 1
    }
}

###############################################################################
# cf parsing
#
proc sandlib::parse {fd result} {
    upvar $result ret

    set ret {}
    set scope global

    set sc [scancontext create]
    scanmatch $sc {^\#} continue
    scanmatch $sc {^\s*\[(.+)\]\s*$} {
	set scope $matchInfo(submatch0)
	keylset ret $scope {}
    }
    scanmatch $sc {^\s*(\S+)\s*=\s*(.*\S)\s*$} {
	keylset ret $scope.$matchInfo(submatch0) $matchInfo(submatch1)
    }

    scanfile $sc $fd
    scancontext delete $sc
}

###############################################################################
# log
#
proc sandlib::log {sw {rest {}}} {
    variable layout
    variable log

    switch $sw {
	open {
	    set logbase [file join $layout(logdir) $rest]
	    foreach {v f} {msg log dbg dbg err err} {
		set log($v) [open [set log(n$v) ${logbase}.$f] a]
		fconfigure $log($v) -buffering line
	    }
	    set log(header) {clock format [clock seconds] -format {%x %X }}
	}
	reopen {
	    if {![info exists log(nmsg)]} return
	    foreach v {msg dbg err} {
		catch {close $log($v)}
		set log($v) [open $log(n$v) a]
		fconfigure $log($v) -buffering line
	    }
	}
	close {
	    foreach f {msg dbg err} {catch {close $log($f)}}
	}
	msg -
	err -
	dbg {
	    if {![string eq $log($sw) {}]} {
		puts $log($sw) "[eval $log(header)]$rest"
	    }
	}
    }
}

###############################################################################
# error handling
#
proc sandlib::perror {} {
    variable onerror
    upvar $onerror err

    set err
}

###############################################################################
# async pipe
#

proc sandlib::popen {cmd {out {}}} {
    set in [open |[concat $cmd 2>@ stdout]]
    if {$out eq {}} {
	set inplace 1
	set out [memchan]
    }

    fconfigure $in -translation binary -encoding binary
    fconfigure $out -translation binary -encoding binary

    log dbg $cmd

    fcopy $in $out -command [namespace code [list lappend state$in]]
    if {![info exists [namespace current]::state$in]} {
	vwait [namespace current]::state$in
    }

    lvarpop [namespace current]::state$in
    set state [set [namespace current]::state$in]
    unset [namespace current]::state$in

    set ret [expr {[info exists inplace] ?
		   [concat [seek $out 0][read $out][close $out]] : {}}]

    if {[catch {close $in} f] || ![lempty $state]} {
	error [expr {$f ne {} ? $f : $state}]
    }

    set ret
}

proc sandlib::chroot {root user cmd {handler {}}} {
    popen [concat /usr/bin/sudo /usr/bin/chrootuid $root $user $cmd] $handler
}

###############################################################################
# file utils
###############################################################################
# a bit resricted version of file join
proc sandlib::safejoin {path name} {
    expr {[file pathtype $name] eq {absolute} ||
	  [llength [file split $name]] != 1 ||
	  [string first \\ $name] != -1 ||
	  [string first { } $name] != -1 ||
	  [string range $name end end] eq {/} ||
	  $name eq {.} || $name eq {..} || $name eq {~} ?
	  [error "wrong name: $name"] :
	  [file join $path $name]}
}

proc sandlib::symlink {src dest} {
    set up {}
    set down {}
    set waterline 0
    set name [file tail $dest]
    foreach \
	s [file split [file dirname $src]] \
	d [file split [file dirname $dest]] {
	    if {!$waterline && [string eq $s $d]} {
		continue
	    } else {
		set waterline 1
	    }
	    
	    if {![string eq $s {}]} {
		set down [file join $down $s]
	    }
	    if {![string eq $d {}]} {
		set up [file join $up ..]
	    }
	}

    link -sym [file join $up $down $name] $dest
}

proc sandlib::readlink file {
    # should be enough
    set maxdepth 64

    while {[string eq [file type $file] link] && [incr maxdepth -1]} {
	set file [file join [file dirname $file] [file readlink $file]]
    }

    if {!$maxdepth} { error "maximum link depth exceeded on $file" }

    set file
}

proc sandlib::digest file {

    set chan [open $file r]
    fconfigure $chan -encoding binary -translation binary
    ::hex -attach [set md [memchan]] -mode encode
    ::md5 -attach $chan -mode write -read-dest $md -read-type channel
    read $chan
    close $chan

    unstack $md
    seek $md 0
    set ret [string tolower [read $md]]
    close $md

    set ret
}

proc sandlib::fileq {one two} {

    expr {[file size $one] == [file size $two] &&
	  [string eq [digest $one] [digest $two]]}
}
