# -*- tcl -*-
# $Id: sandpm.tcl,v 1.181 2004/09/02 22:25:58 me Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval pm {
    namespace import [namespace parent]::*
    namespace export pm

    variable commands [list cget cset stats unmet install names uris source spec build destroy]
    variable counter 0
}

###############################################################################
# opts (keyed list):
# root:   sandbox root dir
# arch:   default build arch
# stable: stable repository
# pocket: pocket repository (optional)
# where `repository' keyed list:
# aptsource: string in apt's `sources.list' format
# binpath:   full path to binary packages on host fs
# srcpath:   full path to source packages on host fs
# pathsubst: list of {apt_path rpm_path} to substitute
###############################################################################
proc pm::pm opts {
    namespace eval [namespace current]::[set id pm[incr [namespace current]::counter]] {
	array set options {}
    }
    upvar [namespace current]::${id}::options options

    set options(defs) {}
    set options(built) {}
 
    if {![keylget opts root root]} {
	error "root not specified"
    }

    if {[file isdirectory $root]} {
	set options(root) $root
    } else {
	error "couldn't initialise on chroot: wrong root $root"
    }
    
    if {![keylget opts arch options(arch)]} {
	error "build arch undefined"
    }
    
    if {[keylget opts stable stable]} {
	set reps stable
    } else {
	error "empty stable repository"
    }
	
    if {[keylget opts pocket pocket]} {
	lappend reps pocket
    }
	
    array set options [list aptsource {} binpath {} srcpath {} pathsubst {}]
    foreach rep $reps {
	foreach k {aptsource binpath srcpath pathsubst} {
	    if {[keylget $rep $k v]} {
		set options($k) [concat $options($k) $v]
	    } else {
		error "$k in $rep undefined"
	    }
	}
    }
    
    set options(aptsource) [lsort -uniq $options(aptsource)]
    set options(binpath) [lsort -uniq $options(binpath)]
    set options(srcpath) [lsort -uniq $options(srcpath)]

    if 0 {
	log dbg --------
	foreach k [array names options] {
	    log dbg "$k\([llength $options($k)]\): '$options($k)'"
	}
	log dbg --------
    }

    rpm:init $id
    apt:init $id

    # XXX creates alias in global namespace
    interp alias {} ::$id {} [namespace current]::_proc $id
    set id
}

proc pm::_proc {id {cmd {}} {args {}}} {
    variable commands

    if {[llength [info level 0]] == 2 } {
	error "wrong \# args: should be \"$id option ?arg arg ...?\""
    }
    if {[lsearch -exact $commands $cmd] == -1} {
	error "wrong command $cmd: should be $commands"
    } else {
	eval [list [namespace current]::$cmd $id] $args
    }
}

#{{{ commands
proc pm::destroy id {
    log msg "\[$id\] destroy"
    namespace delete [namespace current]::$id
    interp alias {} ::$id {}
}

proc pm::cget {id {key {}}} {
    upvar [namespace current]::${id}::options options

    if {[string eq $key {}]} {
	array names options
    } else {
	set options($key)
    }
}

proc pm::cset {id rest} {
    upvar [namespace current]::${id}::options options

    foreach {k v} $rest {
	keylset options(defs) $k $v
    }
}

proc pm::uris {id package {needsubst 1}} {
    upvar [namespace current]::${id}::options options

    apt:uris $id install $package $needsubst
}

# accepts list of binary rpm's
# returns keyed list of root names (as in src.rpm) with svr
proc pm::names {id rest} {
    upvar [namespace current]::${id}::options options

    set rpms [rpm:query $id [list name serial version release sourcerpm] $rest]
    set srpms [apt:uris $id source [string map {^ .} [keylkeys rpms]]]

    foreach p [keylkeys rpms] {
	if {[lsearch -glob $srpms */[keylget rpms $p.sourcerpm]] == -1} {
	    log dbg "no source package found for $p"
	}
    }

    rpm:query $id [list name serial version release] $srpms
}

# installs source package
# returns spec keyed list
proc pm::source {id package {handler {}}} {
    upvar [namespace current]::${id}::options options

    # get full path to source package in repositopy
    # FIXME it won't work if binary packages aren't in repos
    set pkg [apt:uris $id source $package]
    if {[lempty $pkg]} return
    rpm:popen $id [id user] [concat -iv --ignoresize $pkg] $handler

    foreach {{} specfile {}} \
	[intersect3 \
	     [split [rpm:popen $id [id user] [concat -qpl $pkg]] \n] \
	     [glob -nocomplain -tails -directory $options(root)$options(specdir) *]
	] break
    if {[llength $specfile] != 1} {
	error "bogus source rpm: $specfile"
    }
    spec $id [read_file $options(root)$options(specdir)/$specfile]
}

# installs binary package(s) with all dependencies 
proc pm::install {id package {handler {}}} {
    set pkg [apt:uris $id install $package]
    if {[lempty $pkg]} return
    rpm:popen $id root [concat -iv --ignoresize $pkg] $handler
}

# rpm build, returns newly built rpm(s) on success
proc pm::build {id {handler {}}} {
    upvar [namespace current]::${id}::options options

    rpmbuild:popen $id $options(buildcmd) $handler
    set searchpath [list $options(root)$options(rpmdir) \
			$options(root)$options(srcrpmdir)]
    set have [recursive_glob $searchpath *.rpm]
    foreach {missing old built} [intersect3 $options(built) $have] break
    set options(built) $have
    set built
}

# almost noop
proc pm::stats {id package handler} {
    cache:popen $id [concat dotty $package] $handler
}

proc pm::unmet {id handler} {
    cache:popen $id unmet $handler
}

# writes and expands spec with rpm -bE
# returns keyed list of some spec fields
proc pm::spec {id spec} {
    upvar [namespace current]::${id}::options options
 
    set tempspec [file join $options(specdir) $id.[pid]]
    write_file $options(root)$tempspec $spec

    set fd [memchan]
    # -bE with --define is buggy
    catch {
	rpmbuild:popen $id [list -bE $tempspec] $fd
    }

    seek $fd 0

    set spec {}
    set sc [scancontext create]
    scanmatch $sc {(?i)^name:\s*(\S+)\s*$} {
	if {![info exists name]} {
	    set name $matchInfo(submatch0)
	}
    }
    scanmatch $sc {(?i)^version:\s*(\S+)\s*$} {
	if {![info exists ver]} {
	    set ver $matchInfo(submatch0)
	}
    }
    scanmatch $sc {(?i)^release:\s*(\S+)\s*$} {
	if {![info exists rel]} {
	    set rel $matchInfo(submatch0)
	}
    }
    scanmatch $sc {(?i)^%package} {
	set subpackage 1
    }
    scanmatch $sc {(?i)^serial:\s*(\d+)\s*$} {
	if {![info exists serial] && ![info exists subpackage]} {
	    set serial $matchInfo(submatch0)
	}
    }
    scanmatch $sc {(?i)^buildarch:\s*(\S+)\s*$} {
	keylset spec arch $matchInfo(submatch0)
    }
    scanmatch $sc {(?i)^(?:buildrequires|buildprereq):\s*(.+)\s*$} {
	append req $matchInfo(submatch0) { }
    }
    scanmatch $sc {(?i)^(?:source|patch|icon)\d*:\s*(?:(?:.+?://[~/.:\-\+\w\d]+)/)*([.\-\+\w]+)} {
	lappend src $matchInfo(submatch0)
    }
    scanmatch $sc {(?i)^group:\s*(.+\S)\s*$} {
	lappend groups $matchInfo(submatch0)
    }
    scanmatch $sc {(?i)^%changelog.*$} {
	scanmatch $sc {^\*\s*(\S+\s+\S+\s+\S+\s+\S+)\s+([^<]+)(<[^>]+>)\s+(.+)\s*$} {
	    keylset changelog \
		date $matchInfo(submatch0) \
		packagername [string trim $matchInfo(submatch1)] \
		packageremail $matchInfo(submatch2) \
		svr $matchInfo(submatch3)
	    break
	}
    }

    scanfile $sc $fd
    scancontext delete $sc
    seek $fd 0
    set ppspec [read $fd]
    close $fd

    foreach tag {name ver rel} {
	if {[info exists $tag]} {
	    keylset spec $tag [set $tag]
	} else {
	    error "$ppspec\ntag `$tag' not exists in spec"
	}
    }

    if {![info exists groups]} {
	error "$ppspec\ntag `group' not exists in spec"
    }

    foreach group $groups {
	if {![lcontain $options(groups) $group]} {
	    error "$ppspec\nwrong group: $group"
	}
    }

    if {[info exists serial]} {
	keylset spec serial $serial
	set svr $serial:$ver-$rel
    } else {
	keylset spec serial 0
	set svr $ver-$rel
    }

    if {[info exists req]} {
	# special case : foo=1.2.3
	regsub -all {\s*=\s+([^\,\s]+)} $req {=\1 } req
	regsub -all {[<>]=?\s*[^\,\s]+|\([^\)]+\)|,} $req { } req
	regsub -all {\s\s+} $req { } req
	keylset spec req [lsort -uniq [split [string trim $req]]]
    }

    if {[info exists src]} {
	# rpm -bE (was) really buggy !
	keylset spec src [lsort -uniq $src]
    }

    if {![info exists changelog] || ![string eq $svr [keylget changelog svr]]} {
	error "$ppspec\nwrong changelog entry"
    }
    set specfile [file join $options(specdir) $name.spec]
    set options(buildcmd) [list -ba --target $options(arch) $specfile]
    file rename -force -- $options(root)$tempspec $options(root)$specfile

    set spec
}
#}}}
#{{{ apt

proc pm::apt:init id {
    upvar [namespace current]::${id}::options options

    for_recursive_glob f $options(binpath) {apt-0.3* apt-0.5*} {
	lappend apt $f
    }
    for_recursive_glob f $options(binpath) apt-conf* {
	lappend aptconf $f
    }

    if {![info exists apt]} {
	error "no apt package found in $options(binpath)"
    }
    if {[regexp {apt-0\.(\d)*} [file tail [lindex $apt 0]] {} major]} {
	switch $major {
	    5 {
		set these {./etc/apt/apt.conf ./var/*}
		set andthese {./etc/apt/pkgpriorities ./etc/apt/vendors.list}
	    }
	    3 {
		set these {./etc/apt/apt.conf ./etc/apt/rpmpriorities ./var/*}
		set andthese {./etc/apt/vendors.list}
	    }
	    default {
		error "apt major ver $major unsupported"
	    }
	}
    } else {
	error "no apt package found in $options(binpath)"
    }

    set here [pwd]
    try_eval {
	cd $options(root)
	if {![info exists aptconf]} {
	    read_file |[concat /usr/bin/rpm2cpio $apt | /bin/cpio -iumd --quiet $these $andthese]
	} else {
	    read_file |[concat /usr/bin/rpm2cpio $apt | /bin/cpio -iumd --quiet $these]
	    read_file |[concat /usr/bin/rpm2cpio [lindex $aptconf 0] | /bin/cpio -iumd --quiet $andthese]
	}
	if {$major == 3 && ![file exists [file join etc apt pkgpriorities]]} {
	    link -sym rpmpriorities [file join etc apt pkgpriorities]
	}
	foreach s $options(aptsource) {
	    lappend aptsource [concat rpm $s] [concat rpm-src $s]
	}

	write_file [file join etc apt sources.list] [join $aptsource \n]
	file mkdir \
	    [file join var lib apt lists partial] \
	    [file join var cache apt]
    } {
	error "couldn't init apt: [perror]"
    } {
	cd $here
    }

    array set options [list aptget /usr/bin/apt-get aptcache /usr/bin/apt-cache]

    apt:popen $id {-qq update}
}

proc pm::apt:uris {id op package {needsubst 1}} {
    upvar [namespace current]::${id}::options options

    set ret {}
    set notdone 1
    set maxtry 32
    while {$notdone} {
	set fd [memchan]
	set sc [scancontext create]
	set notdone [catch {apt:popen $id [concat --print-uris -qy $op $package] $fd}]
	seek $fd 0
	if {$notdone} {
	    # something goes really wrong
	    if {![incr maxtry -1] || $op eq {source}} {
		set fatal [read $fd]
		scancontext delete $sc
		close $fd
		break
	    }
	    # alternatives
	    scanmatch $sc {(?i)^package\s+(\S+)\s+is\s+a\s+virtual.+:$} {
		set alternatives {}
		set virtual $matchInfo(submatch0)
		scanmatch $sc {(?i)^you\s+should\s+explicitly} {
		    set shortest_len 0
		    set shortest_package {}
		    foreach p $alternatives {
			set l [llength [apt:uris $id $op $p]]
			if {[lempty $shortest_package] || $l <= $shortest_len} {
			    set shortest_package $p
			    set shortest_len $l
			}
		    }
		    if {[set ndx [lsearch -exact $package $virtual]] != -1} {
			lvarpop package $ndx $shortest_package
		    } else {
			lappend package $shortest_package
		    }
		    return
		}
		scanmatch $sc {^\s+(\S+)} {
		    lappend alternatives $matchInfo(submatch0)
		    continue
		}
	    }
	    # obsoletes, provided by other packages
	    scanmatch $sc {(?i)^package\s+(\S+)\s.+exists\s+in\s+the\s+database\.$} {
		set alternatives {}
		set obsolete $matchInfo(submatch0)
		scanmatch $sc {replace\s+it:$} {
		    scanmatch $sc {(?i)^e:.+no\s+installation\s+candidate$} {
			set shortest_len 0
			set shortest_package {}
			foreach p $alternatives {
			    set l [llength [apt:uris $id $op $p]]
			    if {[lempty $shortest_package] || $l < $shortest_len} {
				set shortest_package $p
				set shortest_len $l
			    }
			}
			if {[set ndx [lsearch -exact $package $obsolete]] != -1} {
			    lvarpop package $ndx $shortest_package
			} else {
			    lappend package $shortest_package
			}
			return
		    }
		    scanmatch $sc {^\s+(\S+)} {
			lappend alternatives $matchInfo(submatch0)
			continue
		    }
		}
	    }
	    # fatal
	    scanmatch $sc {unmet\s+dependencies:} {
		scanmatch $sc {(?i)^\s+(\S+):\s+(?:pre){0,1}depends:\s+(\S+)} {
		    set broken $matchInfo(submatch0)
		    set unmet($broken) "due $matchInfo(submatch1)"
		    scanmatch $sc {(?i)^\s+(?:pre){0,1}depends:\s+(\S+)} {
			append unmet($broken) { } $matchInfo(submatch0)
		    }
		}
	    }
	    scanmatch $sc {(?i)^e:.+broken\s+packages\s*$} {
		set fatal [join [array get unmet]]
		set notdone 0
	    }
	    scanmatch $sc {(?i)^e:\s+(couldn't\s+find\s+package.+)$} {
		set fatal $matchInfo(submatch0)
		set notdone 0
	    }
	} else {
	    # ok
	    scanmatch $sc {(?i)^need\s+to\s+get.+\.$} {
		scanmatch $sc {^'file:(.+)'} {
		    append ret $matchInfo(submatch0) { }
		}
	    }
	}
	scanfile $sc $fd
	scancontext delete $sc
	close $fd
    }

    if {[info exists fatal]} {
	error $fatal
    }

    # convert paths host->chroot if needed
    if {$needsubst && [info exists options(pathsubst)]} {
	foreach {str sub} $options(pathsubst) {
	    regsub -all $str $ret $sub ret
	}
    }

    lsort -uniq [split [string trim $ret]]
}

#}}}
#{{{ rpm

proc pm::rpm:init id {
    upvar [namespace current]::${id}::options options

    array set options [list rpm /bin/rpm build /usr/bin/rpmbuild]

    if {![file exists [file join $options(root) bin rpm]]} {
	# XXX TODO find rpm package in pocket, then stable repos
	set rpms [list]
	foreach dir $options(binpath) {
	    foreach f [glob -nocomplain -tails -directory $dir -type f \
			   rpm-4.* rpm-static-4.*] {
		lappend rpms [list $dir $f]
	    }
	}

	while {![lempty $rpms]} {
	    # try to find first most recent rpm-static package, then usual rpm
	    foreach pattern [list -static {}] {
		if {[set found [lindex \
				    [lsort -dictionary -index end \
					 [lsearch -all -inline $rpms *rpm${pattern}-4*]] \
				    end]] ne {}} {
		    set rpm$pattern [eval file join $found]
		}
	    }
	    break
	}

	if {![info exists rpm]} {
	    error "no rpm package found in $options(binpath)"
	}

	set here [pwd]
	try_eval {
	    cd $options(root)
	    file mkdir bin [file join etc rpm] [file join var lib rpm]
	    write_file [file join etc rpm macros.db1] {%_dbapi 3}
	    popen [concat /usr/bin/rpm2cpio $rpm | /bin/cpio -iumd --quiet \
		       [concat ./usr/lib/rpm/GROUPS \
			    [expr {[info exists rpm-static] ? {} :
				   {./bin/rpm ./usr/lib/rpm/rpmrc ./usr/lib/rpm/macros}}]]]
	    if {[info exists rpm-static]} {
		popen [concat /usr/bin/rpm2cpio ${rpm-static} \
			   | /bin/cpio -iumd --quiet ./usr/bin/rpm.static]
		popen [concat /usr/bin/rpm2cpio [file join [file dirname $rpm] lib[file tail $rpm]] \
			   | /bin/cpio -iumd --quiet ./usr/lib/rpm/rpmrc ./usr/lib/rpm/macros]
		link -sym ../usr/bin/rpm.static bin/rpm
	    }

	    rpm:popen $id [id user] --initdb
	} {
	    error "coudn't init rpm under $options(root): [perror]"
	} {
	    cd $here
	}
    }

    # inits rpm layout
    set q [list --define [list _topdir /sandman]]
    set rpmmacros [list specdir sourcedir rpmdir srcrpmdir]
    foreach m $rpmmacros {
	lappend q --eval %_$m
    }
    foreach m $rpmmacros d [split [string trim [rpm:popen $id [id user] $q]] \n] {
	array set options [list $m $d]
    }

    set groups [read_file -nonewline $options(root)/usr/lib/rpm/GROUPS]
    regsub -all { +(\n)} $groups {\1} groups
    set options(groups) [split $groups \n]
}

proc pm::rpm:query {id queryformat package} {
    upvar [namespace current]::${id}::options options

    set packages {}
    set key [lindex $queryformat 0]
    set fields [lrange $queryformat 1 end]

    foreach f $queryformat {
	append qf %\{[string toupper $f]\} { }
    }
    append qf \n

    # rpm spawns in chroot
    regsub -all $options(root) $package {} package

    set fd [memchan]
    rpm:popen $id [id user] [concat -qp --qf [list $qf] $package] $fd
    seek $fd 0

    foreach p [split [string trim [read $fd]] \n] {
	foreach $queryformat [split $p] break;
	foreach f $fields {
	    keylset packages [string map {. ^} [set $key]].$f [set $f]
	}
    }

    close $fd

    set packages
}

#}}}
#{{{ popen variations

proc pm::apt:opts id {
    upvar [namespace current]::${id}::options options

    list \
	-o RPM::RootDir=$options(root) \
	-o Dir::Etc=$options(root)/etc/apt/ \
	-o Dir::State=$options(root)/var/lib/apt \
	-o Dir::Cache=$options(root)/var/cache/apt \
	-o APT::Install::Virtual=true \
	-o APT::Install::VirtualVersion=true \
	-o Debug::pkgProblemResolver=false \
	-o Debug::pkgRPMPM=false
}

proc pm::cache:popen {id cmd {handler {}}} {
    upvar [namespace current]::${id}::options options

    popen [concat $options(aptcache) $cmd [apt:opts $id]] $handler
}

proc pm::apt:popen {id cmd {handler {}}} {
    upvar [namespace current]::${id}::options options

    popen [concat $options(aptget) $cmd [apt:opts $id]] $handler
}

proc pm::rpmbuild:opts id {
    upvar [namespace current]::${id}::options options

    set opts {}
    foreach key [keylkeys options(defs)] {
	lappend opts --define [list $key [keylget options(defs) $key]]
    }
    set opts
}

proc pm::rpmbuild:popen {id cmd {handler {}}} {
    upvar [namespace current]::${id}::options options

    chroot $options(root) [id user] [concat $options(build) [rpmbuild:opts $id] $cmd] $handler
}

proc pm::rpm:popen {id user cmd {handler {}}} {
    upvar [namespace current]::${id}::options options

    chroot $options(root) $user [concat $options(rpm) $cmd] $handler
}

#}}}
