# -*- tcl -*-
# $Id: sandrep.tcl,v 1.409 2004/09/13 10:11:42 me Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval rep {
    namespace import [namespace parent]::*
    namespace import [namespace parent]::pm::*
    namespace import [namespace parent]::cvs::*
    namespace import [namespace parent]::sandbox::*
    namespace export rep control

    upvar [namespace parent]::layout layout
    variable required [list basesystem buildsystem target cvsroot cvsmodule]
    variable optional [list distribution vendor sources]
    variable override [list prefix cachedir ccachedir]

    variable commands [list \
			   pocket getpocket freepocket \
			   box custombox \
			   export iso \
			   addsources getsources \
			   prep build test \
			   query stats \
			   hook destroy]
    variable options {}
    variable counter 0
}

###############################################################################
# init
###############################################################################
proc rep::init {} {
    variable layout
    variable options
    variable required
    variable optional
    variable override

    set cf [file join $layout(etcdir) repos]
    if {[catch {open $cf} f]} {
	error "couldn't open required config file: $f"
    } else {
	parse $f opts
	close $f
    }

    if {[keylget opts global globals]} {
	foreach p $override {
	    if {[keylget globals $p foo]} {
		set layout($p) $foo
	    }
	}
	keyldel opts global
    }

    # cycle over all repos
    foreach repository [keylkeys opts] {
	set rep [keylget opts $repository]
	# check for required parms
	foreach p $required {
	    if {![keylget rep $p foo]} {
		if {[keylget globals $p foo]} {
		    keylset rep $p $foo
		} else {
		    error "required parameter `$p' of `$repository' is missed"
		}
	    }
	}
	
	foreach p $optional {
	    if {![keylget rep $p foo] && [keylget globals $p foo]} {
		keylset rep $p $foo
	    }
	}
	
	keylset rep topdir [file join $layout(prefix) $repository]
	if {[keylget rep stable stable]} {
	    splitrep rep stable
	    keyldel rep stable
	} else {
	    set stable [list_to_rep $repository stable rep]
	    splitrep rep stable
	}

	set current [list_to_rep $repository current rep]
	splitrep rep current
	
	if {![keylget rep sources {}]} {
	    # if sources path is absolute, leave it unchanged
	    keylset rep sources [file join [keylget rep topdir] sources]
	}

	if {[keylget rep ccachedir ccachedir]} {
	    keylset rep ccachepath $ccachedir
	    keyldel rep ccachedir
	} else {
	    keylset rep ccachepath [file join $layout(ccachedir) $repository]
	}

	# cvs stuff
	set cvsargs [list $repository [keylget rep cvsroot] [keylget rep cvsmodule]]
	keylset rep cvs [cvs $cvsargs]
	keylset rep pocketdir [file join [keylget rep topdir] pocket]
	keylset rep isopath [file join [keylget rep topdir] iso]
	keylset rep tmpdir $layout(tmpdir)
	keylset rep cachedir $layout(cachedir)

	keyldel rep topdir
	keyldel rep cvsroot
	keyldel rep cvsmodule
	
	# FIXME check for existance of given repository
	# while doing re-init
	if {[keylget options $repository {}]} {
	    shutdown $repository
	    keyldel options $repository
	}
	keylset options $repository $rep
    }
}

proc rep::dump {} {
    variable options

    foreach r [keylkeys options] {
	log dbg "---- $r ----"
	set rep [keylget options $r]
	foreach k [lsort [keylkeys rep]] {
	    log dbg "$k: '[keylget rep $k]'"
	}
    }

    foreach id [namespace children] {
	log dbg "---- $r ----"
	foreach k [array names $id::options] {
	    log dbg "$k: $id::options($k)"
	}
    }
}

proc rep::shutdown args {
    variable options

    foreach r [expr {[lempty $args] ? [keylkeys options] : $args}] {
	log msg "\[$r\] shutting down"
	if {[keylget options $r.refpm pm] && $pm ne {}} {
	    $pm destroy
	    keyldel options $r.refpm
	}
	if {[keylget options $r.refbox box] && $box ne {}} {
	    $box destroy 0
	    keyldel options $r.refbox
	}
    }
}

###############################################################################
# constructor
###############################################################################
proc rep::rep opts {
    namespace eval [namespace current]::[set id rep[incr [namespace current]::counter]] {
	array set options {}
    }
    upvar [namespace current]::${id}::options options
    upvar $opts private

    if {[string eq $private(repos) {}]} {
        namespace delete [namespace current]::$id
        return [keylkeys [namespace current]::options]
    }

    if {![keylget [namespace current]::options $private(repos) rep]} {
	namespace delete [namespace current]::$id
	error "no such repository `$private(repos)'"
    }

    # options from toplevel
    foreach k [keylkeys rep] {
	set options($k) [keylget rep $k]
    }

    # add private ones
    for_array_keys k private {
	if {![string eq $private($k) {}]} {
	    set options($k) $private($k)
	}
    }

    if {[info exists options(user)] && [info exists options(pass)]} {
	$options(cvs) auth $options(user) $options(pass)
	log msg "\[$id\] authorization passed for $options(user)"
    }

    # check for stable repos
    foreach dl [list src bin base] {
	foreach d $options(stable$dl) {
	    if {![file isdirectory $d] && ![string eq [file type $d] link]} {
		namespace delete [namespace current]::$id
		error "stable $dl directory '$d' not exists!"
	    }
	}
    }

    # check for current repos
    if {![file isdirectory $options(currentpath)] && [catch {
	file mkdir $options(currentsrc) $options(currentbin) $options(currentbase)
	rehash current
    } f]} {
	namespace delete [namespace current]::$id
	error "couldn't create `current' repos skeleton: $f"
    }

    # check for pocket repos
    if {[info exists options(pocket)]} {
	pocket $id
    }

    # check for rest of paths
    foreach path [list ccachepath isopath sources] {
	if {![file isdirectory $options($path)]} {
	    file delete -force -- $options($path)
	    file mkdir $options($path)
	}
    }

    if {![info exists options(mixin)]} {
	set options(mixin) generic
    }

    if {![info exists options(isoprefix)]} {
	set options(isoprefix) \
	    [expr {[info exists options(pocket)] ? $options(pocket) : {cdimage}}]-
    }

    set options(isosuffix) \
	[
	 if {[info exists options(isosuffix)]} {
	     subst -$options(isosuffix)
	 }
	]

    # check for / create reference sandbox
    if {[info exists options(refbox)]} {
	if {$options(refbox) eq {}} {
	    namespace delete [namespace current]::$id
	    error "repository `$private(repos)' is locked, try later"
	}
    } else {
	keylset [namespace current]::options $private(repos).refbox {}
	try_eval {
	    basebox $id [list name $private(repos)]
	    buildbox $id
	    keylset [namespace current]::options \
		$private(repos).refbox $options(box) \
		$private(repos).refpm $options(pm)
	    set options(refbox) $options(box)
	    set options(refpm) $options(pm)
	    array unset options box
	    array unset options pm
	} {
	    destroy $id
	    keyldel [namespace current]::options $private(repos)
	    error "repository `$private(repos)' is broken: removing from list\n[perror]"
	}
    }

    set options(failed) 0
    
    if 0 {
	foreach k [lsort [array names options]] {
	    log dbg "$k: '$options($k)'"
	}
    }

    # XXX creates alias in global namespace
    interp alias {} ::$id {} [namespace current]::_proc $id
    set id
}

proc rep::_proc {id {cmd {}} {args {}}} {
    variable commands

    if {[llength [info level 0]] == 2 } {
	error "wrong \# args: should be \"$id option ?arg arg ...?\""
    }
    if {[lsearch -exact $commands $cmd] == -1} {
	error "wrong command $cmd: should be $commands"
    } else {
	eval [list [namespace current]::$cmd $id] $args
    }
}

###############################################################################
# Commands
###############################################################################
proc rep::destroy id {
    upvar [namespace current]::${id}::options options

    if {[info exists options(mountpoint)]} {
	foreach mntpt $options(mountpoint) {
	    read_file |[concat sudo /bin/umount $mntpt]
	    log msg "\[$id\] umount $mntpt"
	}
    }

    if {[info exists options(pm)]} {
	$options(pm) destroy
    }

    if {[info exists options(box)]} {
	$options(box) destroy [expr {$options(failed) &&
				     [info exists options(keepfailed)] &&
				     $options(keepfailed)}]
    }

    log msg "\[$id\] destroy"

    namespace delete [namespace current]::$id
    catch {interp alias {} ::$id {}}
}

###############################################################################
# creating initial sandbox
#
proc rep::box id {
    upvar [namespace current]::${id}::options options

    basebox $id
    buildbox $id
}

proc rep::custombox {id name} {
    upvar [namespace current]::${id}::options options

    try_eval {
	basebox $id
	$options(pm) install $name $options(handler)
    } {
	error "couldn't create custom box: [perror]"
    }
}

###############################################################################
# preparing package to build in sandbox
#
proc rep::prep {id name} {
    upvar [namespace current]::${id}::options options

    set options(failed) 1

    mixin $id $name
    
    # install explicit buildrequires
    if {[info exists options(buildreq)]} {
	$options(pm) install $options(buildreq) $options(handler)
    }

    set root [$options(box) cget root]

    if {[info exists options(rebuild)] && $options(rebuild)} {
	set spec [$options(pm) source $name]
    } elseif {[$options(cvs) ismodule $name]} {
	# checks out whole module into tmpdir then runs 'make bte'
	# which places all sources into proper place
	# and puts out spec file contents as result
	# at last, spec preprocessed by rpm -bE
	# XXX tmpdir must exist in chroot
	set path [file join sandman BUILD $name]
	set version [expr {[info exists options(pkgver)] ? $options(pkgver) :
			   [info exists options(pockettop)] ? {current} : {tagged}}]
	$options(cvs) module [file join $root $path] $name $version
	set spec [chroot $root [id user] [concat /usr/bin/make bte -sC $path]]
	set spec [$options(pm) spec $spec]
    } else {
	# regular package
	# checks out spec file from cvs, preprocess
	# and places sources (if any) into proper place
	set spec [getspec $id $name]
	set branch [$options(cvs) branchname $name]
	copysrc $id $root[$options(pm) cget sourcedir]
    }

    # install buildrequires if any
    if {[keylget spec req req]} {
	log msg "\[$id\] $name requires: $req"
	$options(pm) install $req $options(handler)
    }

    # some ldconfig magic -- obsolete ?
    chroot $root root /sbin/ldconfig

    # moved here from rep::buildbox
    prep:kernel $id $root

    set options(failed) 0
}

###############################################################################
# build package in sandbox
#
proc rep::build {id name} {
    upvar [namespace current]::${id}::options options

    set options(failed) 1
    log msg "\[$id\] will build: $name"

    mount $id ccache
    # fix root perms
    chroot [$options(box) cget root] root [concat /bin/chown 0.0 . ..]
    set built [$options(pm) build [verbose]]
	
    # save built packages back to repos
    set repos [expr {[info exists options(pockettop)] ? {pocket} : {current}}]
    foreach p $built {
	file copy -force -- $p [expr {[string match *.*src.rpm $p] ?
				      $options(${repos}src) : $options(${repos}bin)}]
	log msg "\[$id\] built: $p"
    }

    rehash $repos

    set options(failed) 0
}

###############################################################################
# generates list of packages to satisfy named package
#
proc rep::export {id name} {
    upvar [namespace current]::${id}::options options

    try_eval {
	set options(box) [sandbox $options(cachedir)]
	set options(pm) [getpm $id stable pocket]
	$options(pm) uris $name 0
    } {
	set options(failed) 1
	error "couldn't get export list: [perror]"
    }
}

###############################################################################
# generates iso image with list of packages to satisfy named package
#
proc rep::iso {id mode names} {
    upvar [namespace current]::${id}::options options

    set interp [interp create -safe]
    try_eval {
	$interp eval [$options(cvs) cget $mode $options(mixin)]
	set installer [$interp eval {set installer}]
	log msg "\[$id\] $installer"
	# validate a bit
	if {[string eq [file pathtype [keylget installer topdir]] absolute]} {
	    error "absolute topdir"
	}
    } {
	error "broken installer options: [perror]"
    } {
	interp delete $interp
    }
    
    foreach {vendor version release} [keylget installer person] break

    set isodest $options(isoprefix)${version}-${release}-$options(mixin)$options(isosuffix)
    set isodest [file join iso ${isodest}.iso]

    set options(failed) 1

    try_eval {
	set options(box) [sandbox $options(cachedir)]
	set options(pm) [getpm $id stable pocket]
    } {
	error "couldn't create iso box: [perror]"
    }

    if {![lempty $names]} {
	set baselist [$options(pm) uris basesystem 0]
	set plist [$options(pm) uris $names 0]
	foreach {{} {} fulllist} [intersect3 $baselist $plist] break
	foreach f $baselist {lappend blist [file tail $f]}
	set baselist [join $blist \n]
	set flist {}
	foreach f $fulllist {lappend flist [file tail $f]}
	set fulllist [join $flist \n]
    }

    # v-v-r
    set root [$options(box) cget root]
    write_file [file join $root .VERSION] \
	"VENDOR=$vendor\nVERSION=$version\nRELEASE=$release"

    try_eval {
	foreach p [split $options(basesystem)] {
	    $options(pm) install $p [verbose]
	}
	foreach p [keylget installer requires] {
	    $options(pm) install $p [verbose]
	}
    } {
	error "couldn't install required packages: [perror]"
    }
    
    # FIXME module-based installer:
    # multiple modules, respect topdir
    set installerdir [file join $root [keylget installer topdir]]

    foreach {module path} [keylget installer module] {
	if {[string eq [file pathtype $path] absolute]} {
	    error "absolute checkout_dir: $path"
	}

	$options(cvs) module [file join $installerdir $path] $module
    }

    if {![lempty $names]} {
	try_eval {
	    # FIXME installer/updater
	    if {[string eq $mode updater]} {
		set rpmdir [file join $installerdir $options(target) RPMS.update]
		set basedir [file join $installerdir $options(target) base]
	    } else {
		set rpmdir [file join $installerdir RPMS]
		set basedir $installerdir
	    }

	    if {![file isdirectory $rpmdir]} {
		file delete -force -- $rpmdir
		file mkdir $rpmdir
	    }

	    if {![file isdirectory $basedir]} {
		file delete -force -- $basedir
		file mkdir $basedir
	    }

	    if {[string eq [keylget installer method] builtin]} {
		write_file [file join $installerdir MANIFEST.base] $baselist
		write_file [file join $installerdir MANIFEST.full] $fulllist
	    }
	    
	    foreach p $plist {
		file copy -force -- [readlink $p] $rpmdir
	    }
	    
	    # get real list of names
	    # not in case of pocket iso build
	    if {![info exists options(pocket)]} {
		set names [$options(pm) names [glob -directory $rpmdir *]]
	    }

	    # make apt indices for updates
	    if {[string eq $mode updater]} {
		popen [concat genpkglist [file join $installerdir $options(target)] update]
	    }

	    set p [glob -nocomplain -directory $options(stablebin) kernel*BOOT*]
	    if {[lempty $p]} {
		puts $options(handler) "warning: no such kernel-BOOT package"
	    } else {
		# RPMS-boot should exist
		file copy -force [lindex $p 0] [file join $installerdir RPMS-boot]
	    }
	} {
	    error "couldn't copy required packages: [perror]"
	}
    }

    mount $id iso

    switch -exact -- [keylget installer method] {
	builtin {
	    # initrd
	    set size 0
	    foreach {s n} [chroot $root root [file join usr lib isosetup initrd]] {
		lappend initrdlist $n
		incr size $s
	    }
	    # FIXME better size calculation
	    set initrdsize [expr $size / 1024 / 1024 + 15]
	    if {$initrdsize > 64} {
		puts $options(handler) "warning: requested initrd size $initrdsize Mb too big, truncated to 64Mb"
		set initrdsize 64
	    }
	    
	    # FIXME initrd fstype hardcoded
	    # TODO got it from installer klist
	    set initrd [sandbox $root \
			    [list \
				 [list size $initrdsize] \
				 [list fstype ext2] \
				 [list fsopts [list -b 1024 -i 1024]]]]
	    try_eval {
		# FIXME linuxrc
		file copy [file join $root usr lib isosetup linuxrc] [$initrd cget root]
		write_file [file join $root tmp initrdlist] [join $initrdlist \n]
		set cmd [list /bin/sh -c [list /bin/tar c -T /tmp/initrdlist |\
					      /bin/tar x -C [file tail [$initrd cget root]]]]
		chroot $root root $cmd $options(handler)
		file delete -force -- [file join $root tmp initrdlist]
	    } {
		error "couldn't create initrd image: [perror]"
	    } {
		set initrd [$initrd destroy 1]
	    }

	    # FIXME lazy destroy
	    if {$initrd eq {}} {
		error "couldn't create initrd image: umount failed"
	    }

	    # gzip & move initrd to isolinux
	    read_file |[concat /bin/gzip -9 $initrd]
	    file rename -force $initrd.gz [file join $root isolinux initrd.img]

	    # iso image
	    set isoopts [list -quiet -o $isodest -iso-level 2 -R \
			     -x /lost+found -x /initrd -m /var/cache/apt/*.bin \
			     -no-emul-boot -boot-load-size 4 -boot-info-table \
			     -b isolinux/isolinux.bin -c isolinux/boot.cat]
	    # exclude mounts
	    foreach m $options(mountpoint) {
		if {[string match $root/* $m]} {
		    regsub $root $m {} x
		    lappend isoopts -x $x
		}
	    }

	    try_eval {
		# fix root perms
		chroot $root root [concat /bin/chown 0.0 . ..]
		chroot $root root [concat /usr/bin/mkisofs $isoopts /] $options(handler)
	    } {
		error "couldn't create iso image: [perror]"
	    }
	}

	external {
	    try_eval {
		if {[keylget installer command cmd] && ![string eq $cmd {}]} {
		    chroot $root root $cmd $options(handler)
		}
	    } {
		error "couldn't create iso image: [perror]"
	    }
	    try_eval {
		file copy -force [file join $installerdir cdimage.iso] \
		    [file join $root $isodest]
	    } {
		error "couldn't copy iso image: [perror]"
	    }
	}
	default {
	    error "method [keylget installer method] not implemented"
	}
    }

    # don't mark unofficial builds
    if {![info exists options(pocket)]} {
	# FIXME delete iso image when mark fails
	$options(cvs) release $release $names
    }

    set options(failed) 0
}

################################################################################
# pocket manipulations
#
proc rep::pocket id {
    upvar [namespace current]::${id}::options options

    set topdir [safejoin $options(pocketdir) $options(pocket)]

    try_eval {
	file mkdir \
	    [file join $topdir SRPMS.pocket] \
	    [file join $topdir $options(target) RPMS.pocket] \
	    [file join $topdir $options(target) base]
    } {
	set err [perror]
	catch {file delete -force $topdir}
	error "couldn't create repository skeleton: $err"
    }

    array set options \
	[list \
	     pocketbase [file join $topdir $options(target) base] \
	     pockettop [list [list $topdir $options(target) pocket]] \
	     pocketbin [file join $topdir $options(target) RPMS.pocket] \
	     pocketsrc [file join $topdir SRPMS.pocket] \
	     pocketpath $topdir]

    rehash pocket
}

proc rep::getpocket {id pocket match} {
    upvar [namespace current]::${id}::options options
    
    set topdir [safejoin $options(pocketdir) $pocket]
    if {![file isdirectory $topdir]} {
	error "no such pocket `$pocket'"
    }

    if {[string eq $match {}]} {
	set match *
    }

    set destdir [file join $::env(TMPDIR) sandd [pid].$id]
    set searchpath [list [file join $topdir $options(target) RPMS.pocket] \
			[file join $topdir SRPMS.pocket]]

    # FIXME this should be rewritten
    set here [pwd]
    try_eval {
	file mkdir $destdir
	for_recursive_glob f $searchpath $match {
	    file copy -force -- $f $destdir
	}
	cd $destdir
	pipe pr pw
	fcntl $pw CLOEXEC 1
	set src [open |[list cpio -o -H crc --quiet <@ $pr]]
	puts $pw [join [glob -nocomplain -types f *] \n]
	close $pw
	fconfigure $src -translation binary
	fconfigure $options(handler) -translation binary
	fcopy $src $options(handler)
    } {
	error "couldn't send pocket $pocket contents: [perror]"
    } {
	cd $here
	file delete -force -- $destdir
	catch {close $src}
	catch {close $pr}
	catch {close $pw}
    }
}

proc rep::freepocket {id pocket} {
    upvar [namespace current]::${id}::options options

    set topdir [safejoin $options(pocketdir) $pocket]

    if {![file isdirectory $topdir]} {
	error "no such pocket `$pocket'"
    } 

    array set options \
	[list \
	     pockettop [list $topdir $options(target) pocket] \
	     pocketbase [file join $topdir $options(target) base] \
	     pocketbin [file join $topdir $options(target) RPMS.pocket] \
	     pocketsrc [file join $topdir SRPMS.pocket] \
	     pocketpath $topdir]

    foreach f [glob -nocomplain -directory $options(pocketsrc) *.rpm] {
	file copy -force -- $f $options(currentsrc)
    }

    foreach f [glob -nocomplain -directory $options(pocketbin) *.rpm] {
	file copy -force -- $f $options(currentbin)
    }

    file delete -force $topdir

    rehash current
}

################################################################################
# repos querying
#

proc rep::query {id query {rest {}}} {
    upvar [namespace current]::${id}::options options

    switch -exact -- $query {
	names {
	    set names [$options(cvs) names]
	    if {[llength $rest]} {
		set all $names
		set names {}
		foreach pattern $rest {
		    while {[set ndx [lsearch $options(match) $all "${pattern}"]] != -1} {
			lappend names [lvarpop all $ndx]
		    }
		}
	    }
	    set names
	}
	versions {
	    set ret {}
	    foreach t [$options(cvs) versions $rest] {
		lappend ret [string map {_ .} [lindex $t 0]]
	    }
	    lsort -dictionary $ret
	}
	stats {
	    $options(refpm) stats $rest $options(handler)
	}
	unmet {
	    $options(refpm) unmet $options(handler)
	}
	default {
	    error "query `$query' isn't known"
	}
    }
}

################################################################################
# statistics
#

proc rep::stats {id query {rest {}}} {
    upvar [namespace current]::${id}::options options
    
    switch -exact -- $query {
	releases {
	    $options(cvs) releases
	}
	names {
	    $options(cvs) names [lindex $options(relid) 0]
	}
	versions {
	    string map {_ .} [$options(cvs) versions $rest [lindex $options(relid) 0]]
	}
	diff {
	    $options(cvs) diff $options(handler) $options(relid)
	}
	log {
	}
	default {
	    error "query `$query isn't known"
	}
    }
}

################################################################################
# package source manipulations
#
################################################################################
# adds sources to named package
#
proc rep::addsources {id name} {
    upvar [namespace current]::${id}::options options

    # check / make lock until commit
    set lockfile [safejoin $options(sources) .\#$name-lock]
    if {[catch {open $lockfile [list WRONLY CREAT EXCL]} f]} {
	if {[string eq [lindex $::errorCode 1] EEXIST]} {
	    # lock exists
	    if {[catch {read_file $lockfile} lockedby]} {
		# can't read lock contents
		error "$name source dir locked: $lockedby"
	    }
	    if {![string eq $lockedby $options(user)]} {
		# locked by another user
		error "$name source dir locked by $lockedby"
	    }
	} else {
	    # smth else: perms etc
	    error "couldn't create lockfile: $f"
	}
    } else {
	# lock obtained
	puts -nonewline $f $options(user)
	close $f
    }

    set here [pwd]
    set destdir [safejoin $options(sources) $name]
    if {![file isdirectory $destdir]} {
	file mkdir $destdir
    }

    try_eval {
	cd $destdir
	fconfigure $options(handler) -translation binary
	set src [open |[list /bin/cpio -iumv --quiet <@ $options(handler) |& cat]]
	fconfigure $src -translation binary
	fcopy $src $options(handler)
    } {
	error "couldn't pull $options(user) data: [perror]"
    } {
	fconfigure $options(handler) -translation auto
	catch {close $src}
	cd $here
    }
}

################################################################################
# returns sources from named package
#
proc rep::getsources {id rest} {
    upvar [namespace current]::${id}::options options

    set name [lvarpop rest]
    if {[lempty $rest]} {
	set rest *
    }

    set options(pm) $options(refpm)
    try_eval {
	mixin $id $name
	set spec [getspec $id $name]
    } {
	error [perror]
    } {
	unset options(pm)
    }

    set branch [$options(cvs) branchname $name]
    # FIXME this should be rewritten
    set here [pwd]
    set destdir [file join $::env(TMPDIR) sandd [pid].$id]
    try_eval {
	file mkdir $destdir
	copysrc $id $destdir $rest
	cd $destdir
	pipe pr pw
	fcntl $pw CLOEXEC 1
	set src [open |[list cpio -o -H crc --quiet <@ $pr]]
	puts $pw [join [glob -nocomplain -types f *] \n]
	close $pw
	fconfigure $src -translation binary
	fconfigure $options(handler) -translation binary
	fcopy $src $options(handler)
    } {
	error "couldn't gather $name's sources: [perror]"
    } {
	cd $here
	file delete -force -- $destdir
	catch {close $src}
	catch {close $pr}
	catch {close $pw}
    }
}

###############################################################################
# Private
###############################################################################
# sandbox manipulations
###############################################################################
proc rep::basebox {id args} {
    upvar [namespace current]::${id}::options options

    # create initial chroot fs
    set options(box) [sandbox $options(cachedir) $args]
    try_eval {
	# inits pm on chroot
	set options(pm) [getpm $id stable pocket]
	# install base system
	foreach p [split $options(basesystem)] {
	    $options(pm) install $p [verbose]
	}
    } {
	set options(failed) 1 
	error "couldn't create initial box: [perror]"
    }	
}

proc rep::buildbox id {
    upvar [namespace current]::${id}::options options
    
    set options(failed) 1

    # common rpm macros
    foreach m {distribution vendor} {
	if {![string eq $options($m) {}]} {
	    $options(pm) cset [list $m $options($m)]
	}
    }
    # packager
    if {[info exists options(user)]} {
	$options(pm) cset [list packager [$options(cvs) user $options(user)]]
    }

    # install build system
    $options(pm) install $options(buildsystem) [verbose]

    set root [$options(box) cget root]
    # create user/group to make rpm happy
    chroot $root root [concat /usr/sbin/groupadd -r -g [id groupid] sandman]
    chroot $root root [concat /usr/sbin/useradd -r -M -g sandman \
				    -d $::env(HOME) -s /dev/null -u [id userid] sandman]
    chroot $root root [concat /bin/mkdir -p $::env(HOME)]
    chroot $root root [concat /bin/chown [id userid]:[id groupid] $::env(HOME)]
    # FIXME the rest of bte-specific rpm macros -- obsolete
    write_file [file join $root$::env(HOME) .rpmmacros] \
	[$options(cvs) cget macros $options(mixin)]
    # make rpm build topdir
    foreach d {BUILD RPMS SOURCES SPECS SRPMS} {
	file mkdir [file join $root sandman $d]
    }

    set options(failed) 0
}

proc rep::prep:kernel {id root} {
    upvar [namespace current]::${id}::options options

    set kver {}
    # ATT: linux-default from glibc-kernheaders < 2.4.21 were broken
    foreach {pat re} [list \
			  [file join usr include linux-* include linux version.h] \
			  [join [list [file join usr include linux] -(\[^/\]+) .+] {}] \
			  [file join usr lib kernel * version.h] \
			  [join [list [file join usr lib kernel] / (\[^/\]+) .+] {}]
		     ] {
	set kver [concat $kver \
		      [split \
			   [regsub -all -line [subst -noc -nob $re] \
				[join [glob -nocomplain -directory $root -tails $pat] \n] {\1}
			   ]
		      ]
		 ]
    }

    if {![llength $kver]} {
	puts $options(handler) "WARNING: kernel headers absent"
	return
    }

    if {[llength $kver] > 1 && [set ndx [lsearch -exact $kver default]] != -1} {
	# set the glibc-kernheaders last alternative
	lappend kver [lvarpop kver $ndx]
    }

    puts $options(handler) "kernel versions found: $kver"
    # if version in options(kernel) isn't supplied, first candidate used
    set kver [expr {![info exists options(kernel)] ? [lindex $kver 0] :
		    [expr {[lcontain $kver $options(kernel)] ? $options(kernel) :
			   [error "no kernel headers of version `$options(kernel)'"]}]}]
    puts $options(handler) "using kernel version: $kver"

    # seems this is ok for both old & new variants
    chroot $root root [concat /usr/bin/env KERNEL_VERSION=$kver \
			   [file join etc rc.d scripts gen_kernel_headers] $kver]
}

###############################################################################
# sources manipulations
###############################################################################
# checking sources
proc rep::checksrc {id spec} {
    upvar [namespace current]::${id}::options options

    if {![keylget spec src sources]} return

    # TODO make this transaction-like:
    # firstly be sure for every source/patch
    # then actually copy/link
    keylget spec name name
    set branch [$options(cvs) branchname $name]
    set searchtop [file join $options(sources) $name]
    set sameserial [file join $searchtop [keylget spec serial]]
    set sameversion [file join $sameserial [keylget spec ver]]
    set samerelease [file join $sameversion [keylget spec rel]]
    set destdir [file join $samerelease $branch]
    set have [glob -nocomplain -directory $destdir -types {l f} *]
    foreach f $sources {
	set src [file join $searchtop $f]
	set dest [file join $destdir $f]
	
	# check for newly added source
	log msg "\[$id\] checking for $f in $searchtop"
	if {[file exists $src]} {
	    # new source appeared
	    if {[lcontain $have $dest]} {
		# we have old one also
		if {[fileq $src [readlink $dest]]} continue
		# new source differs from old one, so check for
		# symlinks in maybe-existing branches
		foreach s [glob -nocomplain -directory $destdir -types l [file join * $f]] {
		    if {[string eq [file normalize [readlink $s]] $dest]} {
			# we have */foo -> ../foo case
			file copy -force $dest $s
		    }
		}
		file rename -force -- $src $dest
		continue
	    } else {
		if {![file isdirectory $destdir]} {
		    file mkdir $destdir
		}
		file rename -force -- $src $dest
		continue
	    }
	}
	# no new source, check for old one
	log msg "\[$id\] checking for $f in $destdir"
	if [lcontain $have $dest] continue
	# find nonexistent source
	findsrc $dest [expr {$branch ne {} ?
			     [list \
				  [file join $sameversion * $branch] \
				  [file join $sameserial * * $branch] \
				  [file join $searchtop * * * $branch] \
				  $samerelease \
				  [file join $sameversion *] \
				  [file join $sameserial * *] \
				  [file join $searchtop * * *]] :
			     [list \
				  [file join $sameversion *] \
				  [file join $sameserial * *] \
				  [file join $searchtop * * *]]}]
	# not found
	lappend notfound $f
    }

    if {[info exists notfound]} {
	error "$notfound not found, try `addsources' first !"
    }

    # cleanup in incoming
    foreach f [glob -nocomplain -directory $searchtop -types f *] {
	file delete -force $f
    }

    # remove lock
    file delete -force -- [file join $options(sources) .\#$name-lock]
}

###############################################################################
# finds named source file around
#
proc rep::findsrc {dest globlist} {
    set name [file tail $dest]
    set destdir [file dirname $dest]
    foreach pat $globlist {
	set have {}
	foreach ele [glob -nocomplain [file join $pat $name]] {
	    lappend have [file split $ele]
	}
	set have [lsort -dictionary $have]
	if {![lempty $have]} {
	    if {![file isdirectory $destdir]} {
		file mkdir $destdir
	    }
	    set src [eval file join [lindex $have end]]
	    symlink $src $dest
	    return -code continue
	}
    }
}
###############################################################################
# lookup in sources XXX
#
proc rep::copysrc {id destdir {match *}} {
    upvar [namespace current]::${id}::options options
    upvar spec sp
    upvar branch br

    if {[keylget sp src sources]} {
	set searchpath [file join $options(sources) [keylget sp name] \
			    [keylget sp serial] [keylget sp ver] [keylget sp rel] $br]
	set have [readdir $searchpath]
	foreach src $sources {
	    if {![string match $match $src]} continue
	    if {[lcontain $have $src]} {
		file copy -force [readlink [file join $searchpath $src]] $destdir
	    } else {
		error "source $src not exists"
	    }
	}
    }
}

###############################################################################
# various various
###############################################################################
# returns list of {name s-v-r} if latter is increased
proc rep::cmpspec {id curspec} {
    upvar [namespace current]::${id}::options options

    keylget curspec name name
    set prevspec [getspec $id $name]

    foreach s {prev cur} {
	keylget ${s}spec serial ${s}_serial
	set ${s}_version [keylget ${s}spec ver]-[keylget ${s}spec rel]
    }

    if {$cur_serial < $prev_serial} {
	error "can't step back from serial `$prev_serial' of `$name'"
    }

    if {$cur_serial > $prev_serial} {
	set newserial 1
    } else {
	if {![string eq $prev_version $cur_version]} {
	    set sorted [lsort -dictionary [list $prev_version $cur_version]]
	    if {![string eq $prev_version [lindex $sorted 0]]} {
		error "can't step back from version `$prev_version' of `$name'"
	    }
	    set newversion 1
	}
    }

    expr {[info exists newversion] || [info exists newserial] ?
	  [expr {[$options(cvs) ismodule $name] ?
		 [list $name $cur_serial-$cur_version] :
		 [list $name $prev_serial-$prev_version]}] : {}}
}

proc rep::getspec {id name} {
    upvar [namespace current]::${id}::options options

    # FIXME better comingfrom
    # log dbg "called from [lindex [info level -1] 0]"
    $options(pm) spec \
	[$options(cvs) spec $name \
	     [expr {[info exists options(pkgver)] ? $options(pkgver) :
		    [expr {[lindex [info level -1] 0] eq "[namespace current]::getsources" ||
			   [lindex [info level -1] 0] eq "cmpspec" ||
			   [info exists options(pockettop)] ? {current} : {tagged}}
		    ]}
	     ]
	]
}

# in: repos list
# out: pm handler
proc rep::getpm {id args} {
    upvar [namespace current]::${id}::options options

    lappend pmargs [list root [$options(box) cget root]] [list arch $options(target)]
    foreach repos $args {
	if {[info exists options(${repos}top)]} {
	    foreach top $options(${repos}top) bin $options(${repos}bin) src $options(${repos}src) {
		lappend aptsource file:$top
		lappend binpath $bin
		lappend srcpath $src
	    }
	    keylset repopts \
		aptsource $aptsource \
		binpath $binpath \
		srcpath $srcpath \
		pathsubst [list $options(${repos}path) $repos]
	    lappend pmargs [list $repos $repopts]
	}
	mount $id $repos
    }

    pm $pmargs
}

proc rep::mixin {id name} {
    upvar [namespace current]::${id}::options options

    set interp [interp create -safe]
    try_eval {
	$interp eval "set name $name"
	$interp eval [$options(cvs) cget release $options(mixin)]
	set common [
		    $interp eval {
			if {[info exists BUILD_ENV(common)]} {
			    set BUILD_ENV(common)
			}
		    }
		   ]
	set mixin [
		   $interp eval {
		       if {[info exists BUILD_ENV($name)]} {
			   set BUILD_ENV($name)
		       }
		   }
		  ]
    } {
	error "mixin: [perror]"
    } {
	interp delete $interp
    }
    
    $options(pm) cset [concat __mixin $options(mixin) $common $mixin]
}

###############################################################################
# if specified mountpath exists, mounts it under chroot
proc rep::mount {id mnt} {
    upvar [namespace current]::${id}::options options

    # XXX be sure you have proper /etc/sudoers
    if {[info exists options(${mnt}path)]} {
	set mntpt [file join [$options(box) cget root] $mnt]
	if {![info exists options(mountpoint)] ||
	    [info exists options(mountpoint)] &&
	    ![lcontain $options(mountpoint) $mntpt]} {
	    file mkdir $mntpt
	    read_file |[concat sudo /bin/mount --bind -o ro $options(${mnt}path) $mntpt]
	    log msg "\[$id\] mount: $options(${mnt}path) -> $mntpt"
	    lappend options(mountpoint) $mntpt
	}
    }
}

###############################################################################
# generates apt hashes
proc rep::rehash repos {
    upvar options opts

    set here [pwd]
    try_eval {
	cd $opts(${repos}base)
	foreach comp $opts(${repos}top) {
	    foreach {path disp rep} $comp break
	    read_file |[concat genpkglist [file join $path $disp] $rep]
	    read_file |[concat gensrclist $path $rep .]
	}
    } {
	error "rehash: [perror]"
    } {
	cd $here
    }
}

###############################################################################
# broken
proc rep::verbose {} {
    upvar options opts

    if {[info exists opts(verbose)]} {
	regsub [namespace current]:: [lindex [info level -1] 0] {} p
	if {[lsearch -exact $opts(verbose) $p] != -1} {
	    set opts(handler)
	}
    }
}

proc rep::splitrep {list str} {
    upvar $list ret $str s

    if {![regexp {file:(/\S+)\s+(\S+)\s+(.+\S)\s*$} $s {} path disp comps]} {
	error "wrong repository `$str'"
    }
    keylset ret ${str}path $path
    keylset ret ${str}base [file join $path $disp base]
    foreach c [split $comps] {
	lappend top [list $path $disp $c]
	lappend bin [file join $path $disp RPMS.$c]
	lappend src [file join $path SRPMS.$c]
    }
    keylset ret ${str}top $top
    keylset ret ${str}bin $bin
    keylset ret ${str}src $src
}

proc rep::list_to_rep {name repos parms} {
    upvar $parms rep

    concat file:[file join [keylget rep topdir] $repos] \
	[keylget rep target] [string tolower $name]
}

###############################################################################
# cvs hooks / callbacks 
###############################################################################
proc rep::hook {id rest} {
    upvar [namespace current]::${id}::options options

    switch -exact -- $options(hook) {
	cvs:commit {
	    $options(cvs) hook commit [list [namespace current]::commit $id] $options(path) $rest
	}
	cvs:tag {
	    $options(cvs) hook tag {} $options(path) $rest
	}
	cvs:dump {
	    $options(cvs) hook dump [list puts $options(handler)] $options(path) $rest
	}
	default {
	    error "no such hook `$options(hook)'"
	}
    }
}

################################################################################
# pre-commit sanity check:
# name, serial-version-release, sources
#
proc rep::commit {id specfile revision} {
    upvar [namespace current]::${id}::options options

    log msg "\[$id\] commit: $specfile $revision"
    
    set ret {}
    set options(pm) $options(refpm)

    try_eval { 
	if {[catch {$options(pm) spec [read_file $specfile]} curspec]} {
	    error "broken `[file tail $specfile]' spec: $curspec"
	}
	
	keylget curspec name name
	if {[file tail $specfile] ne [expr {[$options(cvs) ismodule $name] ?
					    [subst $name.spec] : $name}]} {
	    error "spec file name `[file tail $specfile]' != package name `$name'"
	}
	
	if {![$options(cvs) ismodule $name]} {
	    checksrc $id $curspec
	}
    
	set ret [expr {$revision ne {0} && ![lempty [set tag [cmpspec $id $curspec]]] ?
		       [concat $tag $revision] : {}}]
    } {
	error [perror]
    } {
	unset options(pm)
    }

    set ret
}
