# -*- tcl -*-
# $Id: sandserv.tcl,v 1.100 2003/08/14 11:41:16 lioka Exp $
#
# Copyright (c) 2001,2002,2003 Sergey Bolshakov <sbolshakov@altlinux.ru>
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

namespace eval server {
    package require md5
    namespace import [namespace parent]::*
    namespace import [namespace parent]::task::*

    upvar [namespace parent]::layout layout
    variable revision {$Revision: 1.100 $}
    array set options [list myaddr {} port 65432]
}

#{{{ server
proc server::init {} {
    variable layout
    variable options
    variable revision

    foreach p [array names ::env] {
	if {![lcontain {HOME HOSTNAME PATH TMPDIR} $p]} {
	    catch {unset ::env($p)}
	}
    }

    array set ::env [list \
			 CVS_RSH ssh \
			 PATH /bin:/sbin:/usr/bin:/usr/sbin \
			 TMPDIR $layout(tmpdir) \
			 DURING_INSTALL 1 \
			 BTE_INSTALL 1 \
			 LC_ALL POSIX]

    if {[catch {open [file join $layout(etcdir) sandd]} f]} {
	error "couldn't open required config file: $f"
    }

    parse $f opts
    close $f

    keylget opts global.port options(port)
    keylget opts global.myaddr options(myaddr)
    keylget opts global.debug debug
    keylget opts global.fifo options(ctl)
    keylget opts global.pidfile options(pidfile)
    set options(ctl) [file join $layout(tmpdir) sandd $options(ctl)]

    if {!$debug} {
	set pid [fork]
	if {$pid > 0} {
	    # FIXME wait for child starts
	    exit 0
	}
	
	pidfile $options(pidfile)
	foreach f [file channels] {
	    if {![string match std* $f]} {
		close $f
	    }
	}
	
	set f [open /dev/null a+]
	dup $f stdin
	dup $f stdout
	dup $f stderr
	close $f

	log open sandd
    }

    cd /

    # open control fifo
    file mkdir [file dirname $options(ctl)]
    file delete -force $options(ctl)
    if {[catch {read_file |[list /bin/mkfifo -m0660 $options(ctl)]} f]} {
	error "couldn't create fifo: $f"
    }
    if {[catch {chgrp cvs $options(ctl)} f]} {
	error "couldn't chgrp to cvs: $f"
    }
    set options(ctlfd) [open $options(ctl) r+]
    fconfigure $options(ctlfd) -buffering line -blocking 0
    fileevent $options(ctlfd) readable [namespace current]::ctlaccept

    # set sighandlers
    signal -restart trap {SIGHUP SIGUSR1 SIGUSR2 SIGINT SIGTERM} \
	[list [namespace current]::sighandler %S]

    # opens listening socket
    if {[catch {
	eval socket -server [namespace current]::accept \
	    [expr {$options(myaddr) ne {} ? {-myaddr $options(myaddr)} : {}}] \
	    $options(port)
    } options(servsock)]} {
	error $options(servsock)
    }

    regsub {^.+Revision: ([\.\d]+).*$} $revision {\1} rev
    log msg "sandd ${rev}/$::tcl_version at [info hostname] ($::tcl_platform(os) $::tcl_platform(osVersion))"

    vwait ::forever
}

proc server::pidfile pidfile {
    if {[catch {open $pidfile [list WRONLY CREAT EXCL]} f]} {
	if {[string eq [lindex $::errorCode 1] EEXIST]} {
	    # pidfile exists
	    if {[catch {read_file $pidfile} pid]} {
		error "couldn't read pidfile contents: $pid"
	    }
	    if {[catch {kill 0 $pid} f] && [string eq [lindex $::errorCode 1] ESRCH]} {
		# stale pidfile, trying to reuse
		if {[catch {open $pidfile WRONLY} f]} {
		    error "couldn't open pidfile for writing: $f"
		}
		ftruncate -fileid $f 0
	    } else {
		# seems another process is running
		error "another instance \[$pid\] is running"
	    }
	} else {
	    # smth else: perms etc
	    error "couldn't create pidfile: $f"
	}
    }

    # pidfile opened
    puts -nonewline $f [pid]
    close $f
}

proc server::sighandler sig {
    switch $sig {
	SIGHUP {
	    catch {[namespace current]::log reopen}
	}
	SIGUSR1 {
	    log dbg "reloading configuration"
	    catch {[namespace parent]::rep::init}
	}
	SIGUSR2 {
	    log dbg "dumping configuration"
	    catch {[namespace parent]::rep::dump}
	}
	SIGINT -
	SIGTERM {
	    after idle [namespace current]::shutdown
	}
    }
}

proc server::shutdown {} {
    variable options

    file delete -force $options(ctl)
    catch {close $options(ctlfd)}
    catch {close $options(servsock)}
    foreach sock [file channels sock*] {
	if {[info exists [namespace current]::cl$sock]} {
	    remove $sock "going down"
	} else {
	    catch {close $sock}
	}
    }

    [namespace parent]::rep::shutdown
    shutdown:lazy
}

proc server::shutdown:lazy {} {
    variable options

    if {[llength [after info]]} {
	log dbg "waiting for unfinished jobs ..."
	after 1000 [namespace code shutdown:lazy]
	return
    }

    log msg "clean"
    log close
    file delete -force $options(pidfile)
}
#}}}
#{{{ client handling
proc server::accept {sock ip port} {
    upvar [namespace current]::cl$sock client

    set client(host) $ip
    set client(port) $port
    log msg "incoming connection from $ip:$port"
    fconfigure $sock -buffering line -encoding binary
    fileevent $sock readable [list [namespace current]::client $sock]
}

proc server::client sock {
    upvar [namespace current]::cl$sock client

    if {[gets $sock data] < 0} {
	if {[eof $sock]} {
	    remove $sock "connection closed by peer"
	}
	return
    }

    fileevent $sock readable {}
    fconfigure $sock -buffering full

    catch {task "-handler $sock $data"} clientmsg

    remove $sock $clientmsg
}

proc server::remove {sock {errormsg {}}} {
    upvar [namespace current]::cl$sock client

    if {[string len $errormsg]} {
	catch {puts $sock $errormsg}
	log err $errormsg
    }

    log msg "closing connection to $client(host):$client(port)"
    catch {array unset client}
    catch {close $sock}
}
#}}}
#{{{ hooks
proc server::ctlaccept {} {
    variable options

    if {[gets $options(ctlfd) data] < 0} {
	return
    }

    set data [split $data]
    set client [lvarpop data]
    if {![catch {kill 0 $client} f] || [lindex $::errorCode 1] eq {EPERM}} {
	read_file |[concat /bin/mkfifo -m0660 $options(ctl).$client]
	chgrp cvs $options(ctl).$client
	ctlexec $options(ctl).$client $data
    }
}

proc server::ctlexec {fname data {counter 0}} {
    if {[catch {open $fname {WRONLY NONBLOCK}} id]} {
	if {$counter < 10} {
	    after 500 [list [namespace current]::ctlexec $fname $data [incr counter]]
	} else {
	    file delete -force -- $fname
	}
    } else {
	file delete -force -- $fname
	if {[catch {hook [concat -handler $id $data]} f]} {
	    puts $id $f
	}
	close $id
    }
}
#}}}
